;;;  -*- mode: LISP; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell
;;; Copyright   : (c) 2004 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; Address     : Department of Psychology 
;;;             : Carnegie Mellon University
;;;             : Pittsburgh, PA 15213-3890
;;;             : db30@andrew.cmu.edu
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : attend-and-track.lisp
;;; Version     : 1.0a1
;;; 
;;; Description : Add a new request to the vision module to allow a model to
;;;             : attend to something and start tracking it as a single action.
;;; 
;;; Bugs        : 
;;;
;;; To do       : 
;;; 
;;; ----- History -----
;;; 2015.08.11 Dan [1.0a1]
;;;             : * Added the header info when converting it to ACT-R 7.
;;; 2015.08.18 Dan
;;;             : * Changed randomize-time to randomize-time-ms.
;;; 2015.08.26 Dan
;;;             : * Need to schedule the start-tracking with a priority of 9
;;;             :   because it has to be after the visual chunk gets set, but
;;;             :   before a proc-display sneaks in because that could mess things up.
;;; 2015.09.24 Dan
;;;             : * Add the code to store the request and complete it for the
;;;             :   new request.
;;; 2016.07.20 Dan
;;;             : * Update the code that sets the current marker to use set-current-marker
;;;             :   instead of directly setting the slot.
;;;             : * Updated the code because encoding-complete needs the clof as
;;;             :   a parameter now in addition to the chunk.
;;; 2016.07.21 Dan
;;;             : * Added the clearing of clof to the reset method.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; General Docs:
;;; 
;;; Adds the attend-and-track request to vision which works like move-attention
;;; except that the attended item (if there is one) is also then tracked like
;;; a start-tracking request would initiate.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Public API:
;;;
;;; +visual>
;;;  cmd attend-and-track
;;;  screen-pos <visual-location>
;;;  {scale <scale>}
;;;
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Design Choices:
;;; 
;;; There isn't an easy way to "extend" the vision system so just have to add
;;; modified versions of the functions here plus an around method on the request
;;; to avoid having to redefine that too.
;;;
;;; Not ideal because this means it's harder to update, but easier than changing
;;; the module system, or even just vision, to allow better extenstion capabilities.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; The code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#+:packaged-actr (in-package :act-r)
#+(and :clean-actr (not :packaged-actr) :ALLEGRO-IDE) (in-package :cg-user)
#-(or (not :clean-actr) :packaged-actr :ALLEGRO-IDE) (in-package :cl-user)

(defun reset-vision-module (vis-mod)
  
  (reset-pm-module vis-mod)
  
  (setf (clof vis-mod) nil)
  
  (clrhash (visicon vis-mod))
  (clrhash (found-locs vis-mod))
  
  (setf (current-cursor vis-mod) nil)
  
  (remove-tracking vis-mod)
  (setf (last-scale vis-mod) nil)
  (set-cfs-mth vis-mod :RM-ORIG)
  (setf (synthd-objs vis-mod) (clrhash (synthd-objs vis-mod)))
  (setf (finst-lst vis-mod) nil)
  
  (setf (scene-change vis-mod) nil)
  
  (setf (last-obj vis-mod) nil)
  
  (setf (loc-failure vis-mod) nil)
  (setf (attend-failure vis-mod) nil)
  
  (setf (other-word-chars vis-mod) nil)
  (setf (center-point vis-mod) (vector 0 0))
  
  (setf (process-display-called vis-mod) nil)
  
  
  (chunk-type visual-object screen-pos value status color height width)
  
  (chunk-type (text (:include visual-object)) (text t))
  (chunk-type (empty-space (:include visual-object)) (empty-space t))
  (chunk-type (oval (:include visual-object)) (oval t))
  (chunk-type (cursor (:include visual-object)) (cursor t))
  (chunk-type (line (:include visual-object)) (line t) end1-x end1-y end2-x end2-y)
  (chunk-type (phrase! (:include visual-object)) (phrase! t) objects words colors)
  
  
  (chunk-type visual-location screen-x screen-y distance kind color value height width size)
  (chunk-type (set-visloc-default (:include visual-location)) (set-visloc-default t) type)
  (chunk-type (char-primitive (:include visual-location)) (kind char-primitive) left right)
    
  (chunk-type color color)
  
  (chunk-type vision-command cmd)
  (chunk-type (move-attention (:include vision-command)) (cmd move-attention) screen-pos scale)
  (chunk-type (start-tracking (:include vision-command)) (cmd start-tracking))
  (chunk-type (assign-finst (:include vision-command)) (cmd assign-finst) object location)
  (chunk-type (clear-scene-change (:include vision-command)) (cmd clear-scene-change))
  (chunk-type (clear-all-finsts (:include vision-command)) (cmd clear-all-finsts))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;; Only change is to add this type definition
  
  (chunk-type (attend-and-track (:include move-attention)) (cmd attend-and-track))
  
  
  (dolist (chunk '(lowest highest current current-x current-distance
                   current-y clockwise counterclockwise external
                   internal text box line oval char-primitive 
                   new find-location))
    (unless (chunk-p-fct chunk)
      (define-chunks-fct `((,chunk name ,chunk)))
      (make-chunk-immutable chunk)))

   (dolist (chunk '(black red blue green white magenta yellow
                   cyan dark-green dark-red dark-cyan dark-blue
                   dark-magenta dark-yellow light-gray dark-gray gray
                   pink light-blue purple brown))
    (unless (chunk-p-fct chunk)
      (define-chunks-fct `((,chunk color ,chunk)))
      (make-chunk-immutable chunk)))  
  
  (dolist (chunk '(move-attention assign-finst start-tracking clear-scene-change set-visloc-default clear-all-finsts))
    (unless (chunk-p-fct chunk)
      (define-chunks-fct `((,chunk isa ,chunk)))
      (make-chunk-immutable chunk)))
         
  (setf (default-spec vis-mod) (define-chunk-spec screen-x lowest :attended new)))


(defmethod move-attention-special ((vis-mod vision-module) &key location scale)
  (if (and (eq (exec-s vis-mod) 'BUSY) (not (tracked-obj-last-feat vis-mod)))
    (model-warning "Attention shift requested at ~S while one was already in progress." (mp-time))
    (progn
      (when (tracked-obj-last-feat vis-mod) (remove-tracking vis-mod))
      
      (setf (moving-attention vis-mod) t)
      (clear-attended vis-mod)
      (setf (last-scale vis-mod) scale)
      
      (setf (attend-failure vis-mod) nil)
      
      (set-current-marker vis-mod location)
      
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ;; Only difference is calling encoding-complete-special
      
      (schedule-event-relative (randomize-time-ms (move-attn-latency vis-mod)) 'encoding-complete-special
                               :time-in-ms t :destination :vision :module :vision :params (list location (clof vis-mod) scale) :output 'medium
                               :details (concatenate 'string "Encoding-complete " (symbol-name location) " " (symbol-name scale)))
      
      
      (change-state vis-mod :exec 'BUSY :proc 'BUSY))))


(defmethod encoding-complete-special ((vis-mod vision-module) loc position scale &key (requested t))
  
  (setf (moving-attention vis-mod) nil)
  (change-state vis-mod :exec 'free :proc 'free)
  (set-current-marker vis-mod loc position)
  (complete-request (last-visual-request vis-mod))

  (let ((return-obj (get-obj-at-location vis-mod loc position scale)))
    
    (if return-obj
        (progn
          (set-attended vis-mod (chunk-visicon-entry return-obj))
          (attend-to-object vis-mod return-obj :requested requested)
          
          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
          ;;; Only change is to add this event
          
          (schedule-event-relative 0 'start-tracking :module :vision :output 'medium :destination :vision :priority 9)
     
          return-obj)
      
      (progn
        (clear-attended vis-mod)
        (setf (last-obj vis-mod) nil)
        (set-buffer-failure 'visual :ignore-if-full t :requested requested)
        (setf (attend-failure vis-mod) t)
      
        (schedule-event-now 'no-visual-object-found :maintenance t :module :vision :output 'medium :details "No visual-object found")
        
        nil))))


(defmethod pm-module-request :around ((vis-mod vision-module) buffer-name chunk-spec)
  (if (and (eq buffer-name 'visual)
           (let ((command (chunk-spec-slot-spec chunk-spec 'cmd)))
             (and (= (length command) 1)
                  (eq (spec-slot-value (first command)) 'attend-and-track))))
      
      (let ((sp (verify-single-explicit-value chunk-spec 'screen-pos :vision 'attend-and-track))
            (scale (if (slot-in-chunk-spec-p chunk-spec 'scale)
                       (verify-single-explicit-value chunk-spec 'scale :vision 'attend-and-track)
                     nil)))
        
        (when (visual-lock vis-mod)
        (setf (visual-lock vis-mod) nil)
        (schedule-event-now 'unlock-device :module :vision :destination :device 
                           :priority :min :output nil :maintenance t))
        
        (if (valid-vis-loc-chunk sp vis-mod)
            (progn
              (setf (last-visual-request vis-mod) chunk-spec)
              (schedule-event-now 'move-attention-special 
                                :params (list vis-mod :scale scale :location sp)
                                :details (concatenate 'string "Attend-and-track " (symbol-name sp)" " (symbol-name scale))
                                  :module :vision))
          (progn
            (complete-request chunk-spec)
            (print-warning "screen-pos value ~s in an attend-and-track request was not a valid chunk" sp))))
    
    ;; Only call the original if we're not handling it here
    ;; which means that the visual-lock needs to be handled above.
    
    (call-next-method)))

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#
