;;;  -*- mode: LISP; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell
;;; Copyright   : (c) 2015 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; Address     : Department of Psychology 
;;;             : Carnegie Mellon University
;;;             : Pittsburgh, PA 15213-3890
;;;             : db30@andrew.cmu.edu
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : hold-until.lisp
;;; Version     : 1.0a1
;;; 
;;; Description : Have a motor action which will hold a key until a condition testing
;;;             : the chunk in the visual buffer is true.
;;; 
;;; Bugs        : 
;;;
;;; To do       : [ ] Consider a more general device locking mechanism for productions
;;;             :     like the vision specific one to avoid potential problems in other
;;;             :     modules instead of the explicit tests for procedural being free
;;;             :     and rescheduling as done here to avoid potential jamming.
;;; 
;;; ----- History -----
;;; 2015.08.10 Dan [0.8a1]
;;;             : * Put a real header on this and update it for ACT-R 7 from the
;;;             :   version which ran with the 6.0 space fortress model.
;;;             : * Adapt it to work with the current motor-extension extra
;;;             :   instead of the previous SF hacks.
;;;             : * Test that procedural is not busy before issuing a release
;;;             :   action to avoid jamming with a request in the "firing"
;;;             :   production.  
;;;             : * Instead of separate slots and values to test pass one list
;;;             :   in the spec slot which is a valid chunk-spec to use when
;;;             :   testing the buffer.
;;; 2015.08.27 Dan
;;;             : * Fix a potential issue with a hold-until continuing to check
;;;             :   after some other action (a release-all for example) has 
;;;             :   caused the finger to be released and then later releasing
;;;             :   it either when it's not down or inappropriately.  The quick
;;;             :   fix is to wait for the finger to be down before starting to
;;;             :   check and if it's ever not down after that point stop checking.
;;;             :   There is a better way to handle this because if it didn't 
;;;             :   add the checking test until it put the finger down things
;;;             :   would be cleaner and also work in the event of a cleared
;;;             :   hold-until action, but that would require more integration
;;;             :   with the delayed-punch actions or a lot of after events to
;;;             :   check for things so I'm not doing that now.
;;; 2015.09.24 Dan [1.0a1]
;;;             : * Make sure it completes the requests, but that happens in
;;;             :   the finish movement of the hold-punch action.  That means 
;;;             :   the finger is probably still going to be down when it 
;;;             :   "completes" and potentially creates a problem for utility 
;;;             :   learning with the new mechanism.  For now however these actions
;;;             :   are only a part of the (bad) mine correction code and not
;;;             :   really an issue, but it may need some further consideration
;;;             :   if they're used more generally in the future modeling work.
;;; 2015.09.25 Dan
;;;             : * Removed an ignore of motor from the request function.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; General Docs:
;;; 
;;; Provide an action which will press a finger and then wait for a condition to
;;; be met by the chunk in the visual buffer before releasing it.  Each finger 
;;; can have its own condition for relase.  Relies on the motor-extension extra
;;; for finger and release capabilities.
;;; 
;;; Since it's a hack, for now it doesn't have any delay from condition detection
;;; until the release occurs (other than the timing check interval and possible
;;; lockout because of production firing) and it doesn't require any preparation.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Public API:
;;;
;;; Parameter :hold-until-timing
;;;
;;; Set the period of the test which checks for a hold-until action's conditions
;;; to be satisfied.  Specified in seconds.  There will only be at most one checking
;;; event scheduled which will test all fingers for which a hold-until has been
;;; executed.  It is scheduled with the first one, thus later actions may not be
;;; "in phase" with the checking.
;;;
;;; Motor action: hold-until
;;;
;;; +manual>
;;; cmd hold-until
;;; hand [ left | right ]
;;; finger [ index | middle | ring | pinkie | thumb ]
;;; spec ( <buffer conditions as would be provided for a LHS test> )
;;;
;;; Presses the indicated finger as with a hold-key action and then releases it
;;; when the chunk in the visual buffer satisfies the conditions provided.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Design Choices:
;;; 
;;; Very speculative action used for some space fortress actions to offload the
;;; need to detect a situation to release a key (for example holding until the
;;; speed is "fast enough") because the cost of a production to match and then
;;; request the release can be an issue because of the tight timing in some 
;;; situations.  It may be possible to adjust that other ways in the model, and
;;; avoid this action but for now it makes some things easy...
;;;
;;; It only checks the visual buffer because that's what the model which needs
;;; it is monitoring for actions (the ship in space fortress is being tracked 
;;; and continually updated).  Could imagine a more general version that checked
;;; any buffer for conditions which would eliminate the delayed-punch actions
;;; and replace them with a test of temporal instead, but don't know if that's
;;; really a good idea or not...
;;;
;;; It's definitely too powerful at this point and just being used as a shortcut.
;;; However, there may be some sort of general motor <-> <other module> connections
;;; that would be reasonable to have.
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; The code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#+:packaged-actr (in-package :act-r)
#+(and :clean-actr (not :packaged-actr) :ALLEGRO-IDE) (in-package :cg-user)
#-(or (not :clean-actr) :packaged-actr :ALLEGRO-IDE) (in-package :cl-user)


(defstruct hold-until-module interval event table pressed)

(defun create-hold-until-module (model-name)
  (declare (ignore model-name))
  (make-hold-until-module :table (make-hash-table :size 12 :test #'equalp)
                          :pressed (make-hash-table :size 12 :test #'equalp)))


(defun reset-hold-until-module (module)
  (clrhash (hold-until-module-table module))
  (clrhash (hold-until-module-pressed module))
  (setf (hold-until-module-event module) nil))

(defun handle-hold-until-params (instance param)
   (cond ((consp param)
          (case (car param)
            (:hold-until-timing 
             (setf (hold-until-module-interval instance) (cdr param)))))
         (t
          (case param
            (:hold-until-timing
             (hold-until-module-interval instance))))))


(define-module-fct :hold-until 'nil
  (list
    (define-parameter :hold-until-timing
      :valid-test #'nonneg
      :warning "a number"
      :default-value 0.05
      :documentation "Time between checks of the holding conditions.")
    )
  
  :version "1.0a1"
  :documentation "Module to allow automatic release of keys based on visual checks"
  :creation 'create-hold-until-module
  :reset 'reset-hold-until-module 
  :params 'handle-hold-until-params)

                    
(defun check-hold-until-conditions (module)
  
  (awhen (buffer-read 'visual)
         (maphash (lambda (key value)
                    (let ((pressed (gethash key (hold-until-module-pressed module))))
                      (unless pressed
                        (when (query-buffer (if (eq (first key) 'left) 'manual-left 'manual-right) (list (second key) 'down))
                          (setf pressed t)
                          (setf (gethash key (hold-until-module-pressed module)) t)))
                      (when pressed
                        (if (match-chunk-spec-p it value)
                            (progn
                              (schedule-event-relative 0 'test-for-release :module :motor :destination :motor :params (list (first key) (second key)) :output 'low)
                              (remhash key (hold-until-module-table module))
                              (remhash key (hold-until-module-pressed module)))
                          (when (query-buffer (if (eq (first key) 'left) 'manual-left 'manual-right) (list (second key) 'up))
                              (remhash key (hold-until-module-table module))
                              (remhash key (hold-until-module-pressed module)))))))
           (hold-until-module-table module)))
    
  (when (zerop (hash-table-count (hold-until-module-table module)))
    (delete-event (hold-until-module-event module))
    (setf (hold-until-module-event module) nil)))


(defun test-for-release (motor hand finger)
  (if (query-buffer (if (eq hand 'left) 'manual-left 'manual-right) '(processor free))
    (schedule-event-now 'key-up :module :motor :destination :motor :params (list hand finger (finger-loc-m motor hand finger)) :output nil :priority :max)
    (schedule-event-after-module :motor 'test-for-release :module :motor :destination :motor :params (list hand finger) :output nil)))

(defun handle-hold-until-request (motor chunk-spec)
  (let* ((hand (verify-single-explicit-value chunk-spec 'hand
                                             :motor 'hold-until))
         (finger (verify-single-explicit-value chunk-spec 'finger
                                               :motor 'hold-until))
         (spec (verify-single-explicit-value chunk-spec 'spec
                                             :motor 'hold-until)))
    (if (and hand finger spec (listp spec))
        (setf spec (ignore-errors (define-chunk-spec-fct spec)))
      (setf spec nil))
    
    (unless (find hand '(left right))
      (print-warning "Invalid hand ~s in hold-until request." hand)
      (setf hand nil))
    
    (unless (find finger '(index middle ring pinkie thumb))
      (print-warning "Invalid finger ~s in hold-until request." finger)
      (setf finger nil))
    
    (if (and hand finger (act-r-chunk-spec-p spec))
        (if (finger-down (extension motor) hand finger)
            (progn
              (complete-request chunk-spec)
              (print-warning "The ~S ~S finger is already being held so a hold-until is ignored."  hand finger)             )
          (let  ((hold-module (get-module :hold-until)))
            
            (setf (gethash (list hand finger) (hold-until-module-table hold-module)) spec)
            
            (unless (hold-until-module-event hold-module)
              (setf (hold-until-module-event hold-module) 
                (schedule-periodic-event (hold-until-module-interval hold-module) 'check-hold-until-conditions
                                         :maintenance t :module :hold-until :destination :hold-until :output 'low
                                         :initial-delay (hold-until-module-interval hold-module))))
            
            (schedule-event-now
             'hold-punch
             :destination :motor
             :params (list :hand hand :finger finger :request-spec chunk-spec)
             :module :motor
             :output 'low
             :details (format nil "~a ~a ~a ~a ~a" 'hold-punch :hand hand :finger finger))))
      (progn
        (complete-request chunk-spec)
        (print-warning "Invalid hold-until request no action taken")))))

(extend-manual-requests (hold-until hand finger spec) handle-hold-until-request)



#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#
