const net = require('net');

Experiment = require("./experiment").Experiment;

var Game = require('./game').Game;
var Nav = require('./nav').Nav;
var config = require('./config');

var defaultHost = '';
var defaultPort = 3000;

function json2sexp(json) {
    if (Array.isArray(json)) {
        var ret = '(';
        for (var i=0; i<json.length; i++) {
            if (i>=1) ret += ' ';
            ret += json2sexp(json[i]);
        }
        return ret + ')';
    } else if (typeof json == 'object') {
        var ret = '(';
        var first = true;
        for (var k in json) {
            if (json.hasOwnProperty(k)) {
                if (first) first = false;
                else ret += ' ';
                ret += ':' + k.toString();
                ret += ' ' + json2sexp(json[k]);
            }
        }
        return ret + ')';
    } else if (typeof json === 'string') {
        // special case for keyword symbols
        if (json[0] === ':')
            return json;
        else
            return '"' + json + '"';
    } else if (json === Infinity) {
        return 'inf';
    } else if (json === true) {
        return 'T';
    } else if (json === false) {
        return 'NIL';
    } else if (json === null) {
        console.log('null');
        return 'NIL';
    } else {
        return json.toString();
    }
}

function Client (socket, name, opts) {
    this.sock = socket;
    this.name = name;
    this.lineEnding = opts.lineEnding || '\r\n';
    this.mode = 'config';
    this.commands = [];
    this.logFormat = opts.logFormat || 'native';
    this.exp = new Experiment(opts.dataPath);
    this.configs = opts.configs || {autoturn: new config.AutoturnConfig(),
                                    flappy_autoturn: new config.FlappyAutoturnConfig(),
                                    youturn: new config.YouturnConfig(),
                                    flappy_youturn: new config.FlappyYouturnConfig(),
                                    nav: new config.NavConfig()};
    if (typeof this.configs[opts.condition] === 'undefined') {
        this.condition = 'flappy_youturn';
    } else {
        this.condition = opts.condition;
    }
    this.config = this.configs[this.condition];
    this.gnum = 0;
    this.drawing = true;
}

Client.prototype = {
    // GUI version must implement these functions
    logMsg: function (dest, msg) { console.log(dest, msg); },
    drawingChanged: function () { },
    initGameCanvas: function () { },
    drawScoreScreen: function () { },
    drawGame: function () { },
    updateSubjectID: function (id) { },
    shutdownGUI: function () { },

    greet: function () {
        this.logMsg ('server', 'new connection from ' + this.sock.remoteAddress + ':' + this.sock.remotePort);
        this.logMsg (this.name, 'Config');

        this.sock.write(json2sexp({'screen-type': 'config',
                                   id: this.name,
                                   logFormat: this.logFormat,
                                   condition: this.condition}) + this.lineEnding);
    },

    parseData: function (data) {
        // console.log ('data: ' + data);
        var lines = data.split(/[\n\r]+/);
        // console.log('lines', lines);
        for(var i=0; i<lines.length; i++) {
            lines[i] = lines[i].trim();
            if (lines[i].length > 0) this.commands.push(lines[i]);
        }
    },

    decodeKey: function (str) {
        if (str === 'fire' || str === '32') return 'fire';
        else if (str === 'thrust' || str === '119') return 'thrust';
        else if (str === 'left' || str === '97') return 'left';
        else if (str === 'right' || str === '100') return 'right';
        else return null;
    },

    setDrawing: function(val) {
        if (val === '0') {
            this.drawing = false;
        } else {
            this.drawing = true;
        }
        if (this.mode === 'game') this.drawingChanged();
    },

    startGame: function () {
        this.mode = 'game';
        this.gnum += 1;
        if (this.condition === 'nav') {
            this.exp.setLogFormat('nav');
            this.game = new Nav(this.exp, this.config, this.gnum);
        } else {
            this.exp.setLogFormat('sf');
            this.game = new Game(this.exp, this.config, this.gnum);
        }
        this.game.logFormat = this.logFormat;
        this.initGameCanvas();
        this.game.coreInit();
        this.drawGame();
        this.sock.write (json2sexp(this.game.getWorldStateForModel()) + this.lineEnding);
        this.logMsg (this.name, 'Start Game ' + this.gnum.toString());
    },

    sendScore: function () {
        var val = {'screen-type': 'score',
                   total: this.game.score.pnts,
                   bonus: this.game.bonus,
                   'total-bonus': 0,
                   'raw-pnts': this.game.score.raw_pnts};
        this.sock.write(json2sexp(val) + this.lineEnding);
    },

    handleCommand: function (data) {
        // console.log('cmd', data);
        if (this.mode === 'config') {
            var line = data.split(' ');
            var cmd = line[0];
            var error = null;
            if (cmd === 'id') {
                if (line.length >= 2) {
                    this.updateSubjectID(line[1]);
                    this.name = line[1];
                }
                else error = cmd + ': needs one argument';
            } else if (cmd === 'condition') {
                if (line.length >= 2) {
                    if (line[1] in this.configs) {
                        this.condition = line[1];
                        this.config = this.configs[this.condition];
                    } else {
                        error = cmd + ': unknown condition ' + line[1];
                    }
                } else {
                    error = cmd + ': needs one argument';
                }
            } else if (cmd === 'quit') {
                this.shutdown('client has quit.');
                return;
            } else if (cmd === 'drawing') {
                if (line.length >= 2) this.setDrawing(line[1]);
            } else if (cmd === 'continue') {
                this.logMsg ('server', this.name + ' is starting');
                this.mode = 'game';
                this.exp.id = this.name;
                this.startGame();
                return;
            } else if (cmd === 'logformat') {
                if (line.length >= 2) this.logFormat = line[1];
                else error = cmd + ': needs one argument';
            } else if (cmd === 'config') {
                if (line.length >= 3) {
                    var key = line[1];
                    var val = line[2];
                    if (key === 'display_level') {
                        this.setDrawing(val);
                    } else {
                        error = cmd + ': unknown setting ' + key;
                    }
                } else {
                    error = cmd + ': needs 2 arguments';
                }
            } else {
                error = 'unknown command: ' + cmd;
            }
            this.sock.write(json2sexp({'screen-type': 'config',
                                       id: this.name,
                                       logFormat: this.logFormat,
                                       condition: this.condition,
                                       result: error === null,
                                       error: error}) + this.lineEnding);
        } else if (this.mode === 'game') {
            var line = data.split(' ');
            var cmd = line[0];
            if (cmd === 'quit') {
                this.shutdown('client has quit.');
                return;
            } else if (cmd === 'drawing') {
                if (line.length >= 2) this.setDrawing(line[1]);
            } else if (cmd.toUpperCase() === 'KEYDOWN') {
                if (line.length >= 2) {
                    var k = this.decodeKey(line[1]);
                    if (k) this.game.keyState.press(k);
                }
            } else if (cmd.toUpperCase() === 'KEYUP') {
                if (line.length >= 2) {
                    var k = this.decodeKey(line[1]);
                    if (k) this.game.keyState.release(k);
                }
            } else if (cmd === 'continue') {
                this.game.step_one_tick(33);
                this.drawGame();
                // sock.write (JSON.stringify(game.getWorldState()) + this.lineEnding);
                if (this.game.is_game_over()) {
                    this.logMsg (this.name, 'Game Score ' + this.gnum.toString());
                    this.mode = 'score';
                    this.game.cleanup();
                    this.drawScoreScreen();
                    this.sendScore();
                    return;
                } else {
                    // console.log(this.game.getWorldStateForModel());
                    this.sock.write (json2sexp(this.game.getWorldStateForModel()) + this.lineEnding);
                }
            } else {
                // ignore garbage.
            }
        } else if (this.mode === 'score') {
            var line = data.split(' ');
            var cmd = line[0];
            if (cmd === 'quit') {
                this.shutdown('client has quit.');
                return;
            } else if (cmd === 'continue') {
                this.mode = 'game';
                this.startGame();
                return;
            }
            // this.sendScore();
        }
    },

    handleData: function (data) {
        this.parseData(data);
        for(var i=0; i<this.commands.length; i++) {
            this.handleCommand(this.commands[i]);
        }
        this.commands = [];
    },

    shutdown: function (msg) {
        if (!this.hasShutdown) {
            this.hasShutdown = true;
            this.sock.end();
            this.sock.destroy();
            this.logMsg('server', this.name + ': ' + msg);
            this.shutdownGUI();
        } else {
            console.log('ignore multiple shutdowns.');
        }
    },

    disconnect: function () {
        this.shutdown('client has disconnected.');
    }
};


var clientNum = 0;

function startServer (logFn, clientFn, opts) {
    var port = opts.port || defaultPort;
    net.createServer (function (sock) {
        sock.setEncoding('utf8');
        sock.setNoDelay(true);

        clientNum += 1;
        var client = clientFn(sock, 'model' + clientNum, {lineEnding: opts.lineEnding});
        client.greet();
        sock.on ('data', function (data) { client.handleData(data); });
        sock.on ('error', function (err) {
            if (err.errno === 'ECONNRESET') { }
            else { console.log('socket error', err); }
        });
        sock.on ('close', function () { client.disconnect(); });
    }).listen (port, defaultHost);
    logFn ('server', 'Listening on port ' + port);
}

exports.Client = Client;
exports.startServer = startServer;
